function [abcd, Exx] = abcdwelfare(abcd, beta, Q) 
% ABCDWELFARE: computes discunted and expected geo sum of LQ welfare function
% abcd.EW = trace(Q E x_t x_t')
% where x_{t+1} is state vector of abcd
%
% USAGE: function abcd = abcdwelfare(abcd, Q)
% default: Q = abcd.Q
%
% See also welfareLQ,
% and Sargent & Ljungquist p 47 for derivation (notation map Y=Q, v=W, sigma=Wsigma, C=abcd.B)

%   Coded by  Elmar Mertens, em@elmarmertens.com

%% VERSION INFO
% AUTHOR    : Elmar Mertens
% $DATE     : 07-Jul-2009 11:14:21 $
% $Revision : 1.00 $
% DEVELOPED : 7.8.0.347 (R2009a)
% FILENAME  : abcdwelfare.m

%% default arguments
if nargin < 3 || isempty(Q)
   Q = abcd.Q;
end
if nargin < 2
   if isfield(abcd, 'beta')
      beta = abcd.beta; % notice: discounting is required for conditional welfare, otherwise one ends up with an infinite sume over (undiscouneted) residual variances
   else
      beta = [];
   end
end

%% get VCV of states
A     = abcd.A;
B     = abcd.B;
Exx   = dlyapdoubling(A, B * B');


%% compute unconditional welfare -- henceforth only "scaled PV"
abcd.EW     = trace(Q * Exx);
abcd.Exx   = Exx;

%% decompose unconditional welfare in contributions per shocks
nw          = size(B, 2);
abcd.EWcond = NaN(nw, 1);

for w = 1 : nw
   %    Jw             = diag((1 : nw) == w);
   %    ExxW           = dlyap(A, B * Jw * B');
   exx               = dlyapdoubling(A, B(:,w) * B(:,w)');
   abcd.EWcond(w)    = trace(Q * exx);
end
abcd.EWcond = choppy(abcd.EWcond);
if any(abcd.EWcond  < 0)
   beep
   warning('em:debug','How can (conditional) WELFARE be negative? -- check this out');
   keyboard
   abcd.EWcond(:) = NaN;
   abcd.EW        = NaN;
end

%% now compute conditional welfare 
% follows Sargent & Ljunqvist, (p12 in the 1st edition)
%  w_t = x_t ' W x_t + w

if ~isempty(beta) && beta < 1 % discounting is required, otherwise ill-defined
   A0                   = sqrt(beta) * abcd.A;
   abcd.welfare.W       = dlyapdoubling(A0', Q);
   abcd.welfare.w       = beta / (1 - beta) * ...
      trace(abcd.welfare.W * (abcd.B * abcd.B'));
   
   [eigenstates, lambdas] = eig(abcd.welfare.W);
   lambdas = diag(lambdas);
   lambdas(abs(lambdas) < 1e-5) = 0;
   if any(lambdas < 0)
      %       error('em:msg', 'conditional loss function not p.d.?\nThe eigenvalues are\n%s', mat2sprintf(lambdas))
      warning('em:msg', 'conditional loss function not p.d.?\nThe eigenvalues are\n%s', mat2sprintf(lambdas))
      keyboard
   end
   
   [tmp, ml] = max(lambdas); %#ok<ASGLU>
   abcd.welfare.worstValue = lambdas(ml)^2 + abcd.welfare.w;
   abcd.welfare.worstState = eigenstates(:,ml);
   % notice: eig(a) returns normalized eigenvectors, see MATLAB doc eig
end

