%% BURGUNDY
% $$u_t = F_1 x_t + F_2 \hat{x}_t$$
%
%       $$Ayy^0 Y_{t+1|t} + Ayx^0 X_{t+1|t} = A_{yy} Y_{t|t}  + A_{yx} X_{t|t}  + B_y U_{t|t} $$
%
% USAGE: [hip, pip, F, G, G0, converged, history, iter] = ...
%     burgundypolicyi(F0, G0, beta, Qxx, Qxy, Qyy, Nx, Ny, R, Axx, Axy, Bx, D, Ayy0, Ayy, Ayx, By, Cx, Cu, xNames, yNames, uNames, wNames, zNames, Ayx0)
%
% where Y_{t+1|t} = G0 X_{t+1|t}

%% function header
function [hip, pip, F, G, G0, converged, history, iter] = ...
    burgundypolicyi(F0, G0, beta, Qxx, Qxy, Qyy, Nx, Ny, R, ...
    Axx, Axy, Bx, D, Ayy0, Ayy, Ayx, By, Cx, Cu, ...
    xNames, yNames, uNames, wNames, zNames, Ayx0, ...
    tolerance, maxiter, progressbarFlag)

%% options
if nargin < 26 || isempty(tolerance)
   tolerance   = 1e-10;
end
if nargin < 27 || isempty(maxiter)
   maxiter     = 1e6;
end
if nargin < 28 || isempty(progressbarFlag)
   progressbarFlag = false;
end

useDare = true;

%% prelim
nx = length(xNames);
ny = length(yNames);
nu = length(uNames);
nz = length(zNames); 
% nw = length(wNames);

nS = nx * 2;

Ix = eye(nx);
% Oz = zeros(nz);
if nargin < 25 || isempty(Ayx0)
    Ayx0 = zeros(ny, nx);
end

%% check start values
if isempty(F0)
    F0 = randn(nu, nS);
    warning('em:msg', 'empty start values, using F0 = \n%s', mat2sprintf(F0))
end

%% fixedpoint loop
sqrtbeta    = sqrt(beta);

converged   = false;
iter        = 0;

history.F   = zeros(nu, nS);
history.V   = zeros(nS);
history     = repmat(history, maxiter, 1);

Fnew  = F0;

if isempty(G0)
    F1    = F0(:,1:nx);
    F2    = F0(:,nx+1:end);
    try
       G0    = Gfun(F1 + F2, Axx, Axy, Bx, Ayy0, Ayx0, Ayy, Ayx, By, nx, ny);
    catch  %#ok<CTCH>
       G0 = zeros(ny, nx);
    end
end

G0new = G0;
DD    = D*D';
Oz    = zeros(nz);

while ~converged && iter < maxiter

    iter            = iter + 1;

    %     if iter == 100
    %         progressbarFlag = true;
    %     end
    if progressbarFlag
        progressbar(iter/ maxiter); 
    end
    
    
    F  = Fnew;
    G0 = G0new;
    
   
    F1       = F(:,1:nx);
    F2       = F(:,nx+1:end);
    
    % Kalman
    Atilde  = Axx + Bx * F1;
    Ctilde  = Cx  + Cu * F1;
    if useDare
        Sigma           = dare(Atilde', Ctilde', DD, Oz);
        SigmaZ          = Ctilde * Sigma * Ctilde';
        
        if rcond(SigmaZ) > 1e-10
            K  = Sigma * Ctilde' / SigmaZ;
        else
            K  = Sigma * Ctilde' * pinv(SigmaZ);
        end
        
        convergedKalman = true;
    else
        [Sigma, K, convergedKalman] = abckalmanvfi(Atilde, D, Ctilde, tolerance);
    end

    
    Chat  = Cx  + Cu * (F1 + F2);
    
    
    % y policy: $Y_{t|t} = G_u U_{t|t} + G_x X_{t|t} $$
    Psi0        = Ayy0 * G0 + Ayx0;
    Psi         = Psi0 * Axy - Ayy;
    Gx          = Psi \ (Ayx - Psi0 * Axx);
    Gu          = Psi \ (By - Psi0 * Bx);
    
    Gxu         = Gx + Gu * F1;
    GammaX      = Gxu * K * Cx;
    GammaXhat   = Gxu * (Ix - K * Chat) + Gu * F2;
    GammaU      = Gxu * K * Cu;
    
    % Open Loop: Transition
    ABF   =  Axx + Axy * Gxu + Bx * F1;
    
    AA    = [   Axx + Axy * GammaX  Axy * GammaXhat
                ABF * K * Cx        ABF * (Ix - K * Chat) + (Bx + Axy * Gu) * F2];
            
    BB    = [Bx + Axy * GammaU; ABF * K * Cu];
    
    
    % Open Loop: Loss
    HH = [  Ix              zeros(nx, nx + nu)
            GammaX          GammaXhat           GammaU
            zeros(nu, nS)                       eye(nu)];
    
    QQ = HH' * [Qxx Qxy Nx; Qxy' Qyy Ny; Nx' Ny' R] * HH;
    
    RR = QQ(nS+1:end, nS+1:end);
    NN = QQ(1:nS, nS+1:end);
    %    QQ = QQ(1:nS, 1: nS);
    
    % closed loop: Transfer
    A0 = AA + BB * F;
    if max(abs(eig(sqrtbeta * A0))) > 1
        error('em:msg', 'iter %d: unstable system for dlyapdoubling', iter)
    end
    % closed loop: Loss
    H0 = [eye(nx) zeros(nx); zeros(nx) eye(nx); F];
    Q0 = H0' * QQ * H0;
    % closed loop: Value
    V  = dlyapdoubling(sqrtbeta * A0', Q0);

    history(iter).V = V;
    history(iter).F = F;

    % FOC and new policy
    VV    = V * beta;
    Fnew  = -(RR + BB' * VV * BB) \ (NN' + BB' * VV * AA);
    
    if any(isnan(Fnew(:)))
       warning('em:msg', 'NaN''s in policy function')
       keyboard
    end
    
    G0new = GammaX + GammaXhat + GammaU * (Fnew(:,1:nx) + Fnew(:,nx+1:end));
    
    % check convergence
    dF          = Fnew  - F;
    dG          = G0new - G0;
    converged   = max(abs([dF(:); dG(:)])) < tolerance && convergedKalman;
    
end
if progressbarFlag
    progressbar(1); 
end

%% collect results

F           = Fnew;
G0          = G0new;
history     = history(1:iter);

if converged
   F  = choppy(F, -log10(tolerance));
   G0 = choppy(G0, -log10(tolerance));
   
   [hip, pip, G1, G2] = burgundyrule2abcd(F, G0, beta, Qxx, Qxy, Qyy, Nx, Ny, R, Axx, Axy, Bx, D, Ayy0, Ayx0, Ayy, Ayx, By, Cx, Cu, xNames, yNames, uNames, wNames, zNames, tolerance, useDare);
   G = [G1 G2];
else
    %     deltas = diff(cell2mat({history.F}'));
    %     ndx    = ceil(maxiter / 2): size(deltas, 1);
    %     figure
    %     plotY(abs(deltas(ndx,:)), ndx, [])
    %     set(gcf, 'name', 'Absolute change in most recent steps of policyi algorithm')
    
    error('no convergence after %d iterations', iter)
end

function G = Gfun(F, Axx, Axy, Bx, Ayy0, Ayx0, Ayy, Ayx, By, nx, ny)
% for given policy F, computes a consistent y mapping G

aa = [  eye(nx)         zeros(nx, ny)
        Ayx0            Ayy0
        ];
bb = [  (Axx + Bx * F)  Axy
        (Ayx + By * F)  Ayy
        ];
    
G  = solab(aa, bb, nx);

function [hip, pip, G1, G2, K, Sigma, KC, SigmaStar] = burgundyrule2abcd(F, G, beta, Qxx, Qxy, Qyy, Nx, Ny, R, Axx, Axy, Bx, D, Ayy0, Ayx0, Ayy, Ayx, By, Cx, Cu, xNames, yNames, uNames, wNames, zNames, tolerance, useDare) 

% U =             = F1 X + F2 Xhat
% Y = G X_{t|t}   = G1 X + G2 Xhat

nx = length(xNames);
nu = length(uNames);
nz = length(zNames); %#ok<NASGU>
nw = length(wNames);

% ny = length(yNames);
% nS = nx * 2;

Ix = eye(nx);
Ox = zeros(nx);
% Oz = zeros(nz);

F1    = F(:,1:nx);
F2    = F(:,nx+1:end);

% Kalman
Atilde   = Axx + Bx * F1;
Ctilde   = Cx  + Cu * F1;
if useDare
    Sigma       = dare(Atilde', Ctilde', D * D', zeros(nz));
    K           = Sigma * Ctilde' / (Ctilde * Sigma * Ctilde');
else
    [Sigma, K] = abckalmanvfi(Atilde, D, Ctilde, tolerance);
end
KC          = K * Ctilde;
ImKC        = Ix - KC;
SigmaStar   = ImKC * Sigma;

% Y policy
% $Y_{t|t} = G_u U_{t|t} + G_x X_{t|t} $$
Psi0        = Ayy0 * G + Ayx0;
Psi         = Psi0 * Axy - Ayy;
Gx          = Psi \ (Ayx - Psi0 * Axx);
Gu          = Psi \ (By - Psi0 * Bx);
Gxu         = Gx + Gu * F1;
Chat        = Cx  + Cu * (F1 + F2);

GammaX      = Gxu * K * Cx;
GammaXhat   = Gxu * (Ix - K * Chat) + Gu * F2;
GammaU      = Gxu * K * Cu;

G1 = GammaX    + GammaU * F1;
G2 = GammaXhat + GammaU * F2;

if checkdiff(G1 + G2, G, tolerance * 10)
   warning('em:msg', 'G policy not consistent with expectations, mae = %e', mae(G1 + G2, G))
   disp(G1 + G2)
   disp(G)
end
  

% WRONG:
% G1 = G * KC;
% G2 = G * ImKC;


% Closed Loop Transition
AB1 = Axx  + Axy * G1 + Bx * F1;
AB2 = Axy * G2 + Bx * F2;
A0  = [ AB1         AB2
        AB1 * KC    AB1 * ImKC + AB2];
D0  = [D; zeros(nx, nw)];

% closed loop: Loss
HH = [  Ix  Ox
        G1  G2
        F1  F2
      ];
Q0 = HH' * [Qxx Qxy Nx; Qxy' Qyy Ny; Nx' Ny' R] * HH;

% closed loop value (scaled PV)
% Exx   = dlyapdoubling(A0, D0 * D0');
% EV    = trace(Q0 * Exx);
% 
V     = dlyapdoubling(sqrt(beta) * A0', Q0);

% % closed loop: Value

% v = trace(V * DD0) * beta / (1 - beta);

% % compute unconditionally expected losses
% try
%     if abs(max(abs(eig(A0)) - 1)) < 1e-10
%         warning('em:msg', 'removing unit root for welfare calculation')
%         % check whether no unit roots in loss function
%         [vLoss, dLoss] = eig(V);
%         L = vLoss * sqrt(dLoss); % policy function of "target variables" in loss (or a choleski rotation thereof)
%         if max(abs(eig(L * A0))) >= 1
%             warning('em:msg', 'PLEASE CHECK: it seems there are unit roots in the loss function ...')
%         end
%         Abar    = A0^1e8;
%         A0stat  = (eye(size(Abar)) - Abar) * A0;
%         D0stat  = (eye(size(Abar)) - Abar) * D0;
%         Exx     = dlyapdoubling(A0stat, D0stat * D0stat');
%     else
%         Exx   = dlyapdoubling(A0, DD0);
%     end
%     EV    = v + trace(V * Exx);
% catch %#ok<CTCH>
%     EV  = NaN;
%     warning('em:msg', 'Something did not work with dlyapdoubling (though unit roots had been removed)')
% end


%% Hidden Info State Space

hip.A   = A0;
hip.B   = D0;
hip.Cx  = [F; G1 G2; KC ImKC; Ox Ix; F1 -F1; zeros(nu,nx) F1 + F2; ImKC, -ImKC; (Cx + Cu * F1) Cu * F2];
hip.Q   = Q0;
hip.V   = V;

% names
xHatNames  = cellfun(@(x) sprintf('%s_{t|t-1}', x), xNames, 'uniformoutput', false);
xProjNames = cellfun(@(x) sprintf('%s_{t|t}', x), xNames, 'uniformoutput', false);
% xInnoNames = cellfun(@(x) sprintf('%s_{t}-%s_{t|t-1}', x, x), xNames, 'uniformoutput', false);
xHideNames = cellfun(@(x) sprintf('%s_{t}-%s_{t|t}', x, x), xNames, 'uniformoutput', false);
uInnoNames = cellfun(@(x) sprintf('%s_{t}-%s_{t|t-1}', x, x), uNames, 'uniformoutput', false);
uProjNames = cellfun(@(x) sprintf('%s_{t|t}', x), uNames, 'uniformoutput', false);
uHatNames  = cellfun(@(x) sprintf('%s_{t|t-1}', x), uNames, 'uniformoutput', false);

hip.Snames = xNames;
hip.Xnames = cat(2, xNames, xHatNames);
hip.Ynames = cat(2, uNames, yNames, xProjNames, xHatNames, uInnoNames, uHatNames, xHideNames, zNames);
hip.Wnames = wNames;

hip            = abcddims(hip);
if all(abs(eig(hip.A)) < 1)
    hip            = abcdwelfare(hip, beta);
end
hip.K          = K;
hip.Sigma      = Sigma;
hip.SigmaStar  = SigmaStar;

%% public Info State Space
% as in paper

pip.Xnames = xHatNames;
pip.Ynames = cat(2, uProjNames, yNames);
pip.Wnames = zNames;

ssq  = cholcov(Sigma)';
cssq = Ctilde * ssq;
pip.VarZ = cssq * cssq';
% pip.VarZ = (pip.VarZ + pip.VarZ') / 2;
cholVarZ = cholcov(pip.VarZ)';

pip.A    = Axx + Axy * (G1 + G2) + Bx * (F1 + F2);
pip.B    = AB1 * K * cholVarZ;

pip.C    = [F1 + F2; G1 + G2];
pip.D    = [F1; G1] * K * cholVarZ;

pip      = abcddims(pip);
