%% MARKOV POLICES UNDER FULL INFORMATION
% [abcd, F, G, V, v, history, iter] = markovfullinfopolicyi(F0, G0, beta, Qxx, Qxy, Qyy, Nx, Ny, R, Axx, Axy, Bx, D, Ayy0, Ayy, Ayx, By, xNames, yNames, uNames, wNames, Ayx0)

%% function header
function [abcd, F, G, V, v, history, iter] = markovfullinfopolicyi(F0, G0, beta, Qxx, Qxy, Qyy, Nx, Ny, R, Axx, Axy, Bx, D, Ayy0, Ayy, Ayx, By, xNames, yNames, uNames, wNames, Ayx0, maxiter, tolerance)

%% options and preliminaries
if ~exist('maxiter', 'var')
   maxiter           = 1e4;
end
if ~exist('tolerance', 'var')
   tolerance         = 1e-10;
end


%% names
% xHatNames   = cellfun(@(x) sprintf('\\hat{X}:%s', x), xNames, 'uniformoutput', false);

%% prelim
nx = length(xNames);
ny = length(yNames);
nu = length(uNames);
nw = length(wNames); %#ok<NASGU>

Ix = eye(nx);

if nargin < 22 || isempty(Ayx0)
    Ayx0 = zeros(ny, nx);
end


%% check start values
if isempty(F0)
    F0 = randn(nu, nx); %    F0 = [ones(nu, nx) zeros(nu, nx)];
    warning('em:msg', 'empty start values, using F0 = \n%s', mat2sprintf(F0))
end
if isempty(G0)
    G0 = Gfun(F0, Axx, Axy, Bx, Ayy0, Ayx0, Ayy, Ayx, By, nx, ny);
end

%% fixedpoint loop
converged   = false;
iter        = 1;
history.F   = zeros(nu, nx);
history.V   = zeros(nx);
history     = repmat(history, maxiter, 1);

sqrtbeta    = sqrt(beta);
Q           = [Qxx Qxy Nx; Qxy' Qyy Ny; Nx' Ny' R];

Fnew        = F0;
Gnew        = G0;

while ~converged && iter < maxiter
    
    F = Fnew;
    G = Gnew;
    
    history(iter).F	= F;
    
    % Open Loop: y policy $Y_{t} = G_u U_{t} + G_x X_{t}$$
    Psi0        = Ayy0 * G + Ayx0;
    Psi         = Psi0 * Axy - Ayy;
    Gx          = Psi \ (Ayx - Psi0 * Axx);
    Gu          = Psi \ (By - Psi0 * Bx);
        
    % Open Loop: Loss
    HH = [Ix zeros(nx, nu); Gx Gu; zeros(nu, nx) eye(nu)];
    qq = HH' * Q * HH;     %    QQ = qq(1:nx, 1:nx);
    
    RR = qq(nx+1:end, nx+1:end);
    NN = qq(1:nx, nx+1:end);
    
    % Closed Loop: Transition
    AA    = Axx + Axy * Gx;
    BB    = Bx + Axy * Gu;
    A0    = AA + BB * F;
    
    % Closed Loop: Loss
    H0 = [Ix; G; F]; % using G, since this is for next period's loss function
    Q0 = H0' * Q * H0;
    
    % Closed Loop: Value
    V     = dlyapdoubling(sqrtbeta * A0', Q0);
    VV    = V * beta;
    
    history(iter).V	= V;


    if any(isnan(VV(:)))
        error('something went wrong with dlyapdoubling')
    end
    
    % FOC
    Fnew  = -(RR + BB' * VV * BB) \ (NN' + BB' * VV * AA);
    Gnew  = Gx + Gu * Fnew;
    
    % check convergence
    dF          = F - Fnew;
    dG          = G - Gnew;
    converged   = max(abs([dF(:); dG(:)])) < tolerance;
    
    % prep next round
    iter = iter + 1;
    
end
history = history(1:iter);

if ~converged
    
    dispcalc(F0)
    dispcalc(F)
    error('reached maximum loops but without convergence: current accuracy after %d iterations is %8.4e', iter, max(abs(dF(:))))
    
else % closed loop stuff
    
    % Closed Loop: Transition

    A0    = AA + BB * F;
    
    % Closed Loop: Loss
    H0 = [Ix
        G
        F];
    Q0 = H0' * Q * H0;
    
    % Closed Loop: Value
    V = dlyapdoubling(sqrtbeta * A0', Q0);
    v = trace(V * (D * D')) * beta / (1 - beta);
    
end

abcd.converged = converged;

% state space
abcd.A   = A0;
abcd.B   = D;
abcd.Cx  = [F; G; Ix];
abcd     = abcddims(abcd);

abcd.Xnames = xNames;
abcd.Ynames = cat(2, uNames, yNames, xNames);
abcd.Wnames = wNames;


abcd.Q      = Q0;
abcd        = abcdwelfare(abcd, beta);

function G = Gfun(F, Axx, Axy, Bx, Ayy0, Ayx0, Ayy, Ayx, By, nx, ny)
% for given policy F, computes a consistent y mapping G

aa = [  eye(nx)         zeros(nx, ny)
        Ayx0            Ayy0
        ];
bb = [  (Axx + Bx * F)  Axy
        (Ayx + By * F)  Ayy
        ];
    
G  = solab(aa, bb, nx);