%% EHL model, $\Delta a$ shocks and RW inflation target plus 201-xbar shocks
%  indexation of prices and wages to lagged inflation and lagged beliefs
%
% There are several shocks:   1) iid shock to tech growth (natural wage growth) 
%                             2) one real rate shock (AR 1)
%                             3) RW shock to inflation target
%                             4) 201 shocks to xbar
%
% $$L_t = \alpha_p (\pi_p - \bar\pi)^2 + \alpha_w (\pi_w- \bar\pi)^2 + \alpha_x (x-xbar)^2$$
%
% $$\Delta a_{t} = \mu_{t-1} + e_t$$
% 
% $$\mu_t = \rho \mu_{t-1} + \epsilon_{t+1}$$
%
% $$\bar \pi_t = \bar \pi_{t-1} + \epsilon^\pi_{t+1} $$
%
% $$\hat \pi_t = \pi_t - \bar\pi_{t|t} $$
%
% $$\tilde\omega_t = \tilde\omega_{t-1} + \pi_t^w - \pi_t^p - \psi_{wa}\Delta a_t$$
%
% $$ X_t = [\Delta a_t; \mu_t; \bar\pi_t; \tau_t, \epsilon_t; \tilde\omega_{t-1}] $$ 
%
% $$ Z_t = [\Delta a_t; i_t] $$

% indexation to lagged inflation at rate gamma  and (1-gamma) to CURRENT beliefs

eval(string4ini('gamma', 1));

% persistence parameters
eval(string4ini('rho_x', .9));
eval(string4ini('rho_a', .25))

% variance of inflation target
eval(string4ini('sig_pibar', -.5));
eval(string4ini('sig_n', 1));

% variances of output gap targets
% eval(string4ini('varx', 2));
% eval(string4ini('xPermShare', .5));
% sig_xtau   = sqrt(xPermShare * varx * (1 - rho_x^2));
% sig_xnoise = sqrt((1 - xPermShare) * varx);
% checkdiff(varx, sig_xtau^2 / (1 - rho_x^2) + sig_xnoise^2);

% variances of technology shocks
eval(string4ini('vara', 2));
eval(string4ini('aPermShare', .5));
sig_atau   = sqrt(aPermShare * vara * (1 - rho_a^2));
sig_anoise = sqrt((1 - aPermShare) * vara);
checkdiff(vara, sig_atau^2 / (1 - rho_a^2) + sig_anoise^2);



modelname = strrep(mfilename, 'setup', '');
setpref('LQmodels', 'setupfile', mfilename);


SumAlphas = alpha_x + alpha_p + alpha_w;
alpha_x   = alpha_x / SumAlphas;
alpha_p   = alpha_p / SumAlphas;
alpha_w   = alpha_w / SumAlphas;

display(alpha_x)
display(alpha_p)
display(alpha_w)

%% setup model matrices
xNames = {'\Delta a', '\mu', '\pi^\ast', '\tau_x', '\epsilon_x', 'n', ...
    '\omega(-1)', '\pi^p(-1)', '\pi^w(-1)'}; 
yNames = {'\hat\pi^p', '\hat\pi^w', '\omega', '\pi^p', '\pi^w', 'x', 'r'};
uNames = {'i'};
zNames = {'\Delta a', 's', 'i'}; 
wNames = xNames(1:6);
wNames{1} = '\epsilon^a';

nx = length(xNames);
ny = length(yNames);
nu = length(uNames);
nz = length(zNames);
nw = length(wNames);

Ix = eye(nx);

% Loss
Qxx          = zeros(nx); 
Qxx(3,3)     = alpha_p + alpha_w;
Qxx(4:5,4:5) = alpha_x;


Qxy         = zeros(nx,ny);
Qxy(3,4)    = -alpha_p;
Qxy(3,5)    = -alpha_w;
Qxy(4:5,6)  = -alpha_x;

Qyy   = diag([0, 0, 0, alpha_p, alpha_w, alpha_x,0]);

Nx          = zeros(nx,nu);
Ny          = zeros(ny,nu);
R           = 0;

% x-cons
Axx         = diag([0, rho_a, 1, rho_x, 0, 0, 0, 0, 0]);
Axx(1,2)    = 1; % lagged mu on da
Axy         = zeros(nx, ny);
Axy(7,3)    = 1; % lagged omega
Axy(8,4)    = 1; % lagged price inflation
Axy(9,5)    = 1; % lagged wage inflation


Bx          = zeros(nx,nu);
D           = diag([sig_anoise, sig_atau, sig_pibar, sig_xtau, sig_xnoise, ...
    sig_n, 0, 0, 0]);
D           = D(:,1:nw);

% y-cons: PC-p, PC-w, def omega, def pihat-p, def pihat-w, IS, Fisher,

Ayy0  = [   -beta   0       0    0  0 0         0   
            0       -beta   0    0  0 0         0   
            0       0       0    0  0 0         0   
            0       0       0    0  0 0         0   
            0       0       0    0  0 0         0   
            0       0       0    1  0 sigma     0   
            0       0       0    1  0 0         0   
        ];
    
Ayy   = [   -1       0        lambda_p   0  0 kappa_p   0   
            0       -1        -lambda_w  0  0 kappa_w   0   
            0        0       -1         -1  1 0         0   
            1        0        0         -1  0 0         0   
            0        1        0          0 -1 0         0    
            0        0        0          0  0 sigma     0    
            0        0        0          0  0 0         -1    
         ];

Ayx      = zeros(ny, nx);
% omega loadings
Ayx(3,1) = -psi_wa; % on productivity growth
Ayx(3,7) = 1;       % on lagged omega
% indexation 
Ayx(4,8) = gamma;
Ayx(5,9) = gamma;
Ayx(4:5,3) = 1-gamma;
% real rate
Ayx(6,2) = -sigma;


% i-\pi^\ast
By    = [0; 0; 0; 0; 0;1;1];

% Z-cons
Cx    = [1 0 0 0 0 0 0 0 0; 0 0 1 0 0 1 0 0 0; 0 0 0 0 0 0 0 0 0];
Cu    = [0; 0; 1];

%% dimension checks
dimchk(Qxx, nx, nx);
dimchk(Qxy, nx, ny);
dimchk(Qyy, ny, ny);
dimchk(Nx, nx, nu);
dimchk(Ny, ny, nu);
dimchk(R, nu, nu);
dimchk(Axx, nx, nx);
dimchk(Axy, nx, ny);
dimchk(Bx, nx, nu);
dimchk(D, nx, nw);
dimchk(Ayy0, ny, ny);
dimchk(Ayy, ny, ny);
dimchk(Ayx, ny, nx);
dimchk(By, ny, nu);
dimchk(Cx, nz, nx);
dimchk(Cu, nz, nu);

%% check posdef loss
if any(eig([Qxx Qxy Nx; Qxy' Qyy Ny; Nx' Ny' R]) + 1e-10 < 0)
    error('Loss function is not positive definite. The eigenvalues are \n %s', sprintf('\t%20.12f\n', eig([Qxx Qxy Nx; Qxy' Qyy Ny; Nx' Ny' R]))) %#ok<SPERR>
end